// edited by Hongyi
// buffer is going to encoding for send/recieve
#include "analysisPF.h"

using namespace std;

#define this_dp module_SSA->dp

/**********************************************************
 *********************************************************/


/**********************************************************
  *  print the values of a vector in CFG node for debug
 *********************************************************/

int is_interested(FilteredCFGNode<IsDFAFilter>& cfgNode) {
  SgNode* sgNode = cfgNode.getNode();
    //edited by Hongyi 
   // we are only interested in expressions functioncalls directives statements ,but not variable and value 
  // 1 :for some Omp protected statements either, such as atomic, single, master  
  // 2: verify hybrid model
   return (!(isSgVarRefExp(sgNode)) && !(isSgIntVal(sgNode)) && !(isSgReturnStmt(sgNode)));
}

template <typename T >
void printCFGVector(std::vector< T > worklist) {
  int pos = 0;
  // since openmp cfg only supports several directives.
  // not sure do the analysis inside SgOmpStatement
  std::cout << "\n!! Worklist in CFG node lists: " ;
  for (typename std::vector<T >::iterator i = worklist.begin(); i != worklist.end(); ++i, pos++) {  
    T node = *i;
    std::cout << " - visit - now:  " << node.toStringForDebugging();
  }
  std::cout << std::endl;
}

/**********************************************************
 *  This is the main symbolic analysis function at each function
 *  It determines what should happen at each DFG node
 *********************************************************/

void print_map (map<string, int>& m) {
  for (map<string, int>::iterator ii = m.begin(); ii != m.end(); ii++) {
    cout << ii->first << " = " << ii->second << endl;
  }
}

void print_order (map<string, list<set<Expr> > > &m, DP* dp) {
  for (map<string, list<set<Expr> > >::iterator ii = m.begin(); ii != m.end(); ii++) {
    cout << "[ ";
    for (list<set<Expr> >::iterator jj = ii->second.begin(); jj != ii->second.end(); jj++) {
      cout << "{";
      for (set<Expr>::iterator kk = (*jj).begin(); kk != (*jj).end(); kk++) {
	dp->pp_expr(*kk);
	cout << " ";
      }
      cout << "}";
    }
    cout << " ]" << endl;
  }
}

pair<Expr,Expr> AnalysisPF::merge_branches(Expr p, VarNumMap& l_map, VarNumMap& r_map, VarNumMap& result_map) {

  cout << "l_map: " << endl;
  print_map(l_map);
  cout << "r_map: " << endl;
  print_map(r_map);

  int size = l_map.size() + r_map.size();

  Expr l_args[size];
  int l_index = 0;
  Expr r_args[size];
  int r_index = 0;

  // check l_map first
  for (VarNumMap::iterator ii = l_map.begin(); ii != l_map.end(); ii++) { // for all variables
    string v = ii->first;
    VarNumMap::iterator jj = r_map.find(v);
    if (jj != r_map.end()) {   // in both maps

      //update the result map so that each local variable has its maximum SSA value
      result_map[v] = ii->second > jj->second ? ii->second : jj->second;  
      //Note: we don't need to consider the equal case
      if (ii->second < jj->second) {
	Expr v1 = module_SSA->mk_local_var(v, r_map);
	Expr v2 = module_SSA->mk_local_var(v, l_map);
	l_args[l_index++] = this_dp->mk_eq(v1, v2);
      }
    }
    else {
      result_map[v] = ii->second;
      
      Expr v1 = module_SSA->mk_local_var(v, l_map);
      Expr v2 = module_SSA->mk_local_var(v, r_map);
      r_args[r_index++] = this_dp->mk_eq(v1, v2);
    }
  }
  Expr l_exp = this_dp->mk_and(l_args, l_index);
  
  // check l_map first
  for (VarNumMap::iterator ii = r_map.begin(); ii != r_map.end(); ii++) {
    string v = ii->first;
    cout << "v = " << v << endl;
    VarNumMap::iterator jj = l_map.find(v);
    if (jj != l_map.end()) {   // in both maps

      //update the result map so that each local variable has its maximum SSA value
      result_map[v] = ii->second > jj->second ? ii->second : jj->second;  
      //Note: we don't need to consider the equal case
      if (ii->second < jj->second) {
	Expr v1 = module_SSA->mk_local_var(v, l_map);
	Expr v2 = module_SSA->mk_local_var(v, r_map);
	r_args[r_index++] = this_dp->mk_eq(v1, v2);
      }
    }
    else {
      result_map[v] = ii->second;
       Expr v1 = module_SSA->mk_local_var(v, r_map);
       Expr v2 = module_SSA->mk_local_var(v, l_map);
       l_args[l_index++] = this_dp->mk_eq(v1, v2);      
    }
  }

  cout << "result_map: " << endl;
  print_map(result_map);

  Expr r_exp = this_dp->mk_and(r_args, r_index);

  return make_pair(l_exp, r_exp);

//   Expr exp;
//   if (l_index == 0 && r_index == 0)
//     return this_dp->t_expr;
//   else
//     return this_dp->mk_ite(p, l_exp, r_exp);
}

void AnalysisPF::makeSharedVarOrder(ModelPF::VarOrderMap& order, list<Expr>& constraints) {  // order the schedule ids

  if (Config::use_POR) 
    return;

  vector<Expr> ev;
  for (ModelPF::VarOrderMap::iterator ll = order.begin(); ll != order.end(); ll++) {  // for all variables
    list< set<Expr> >::iterator kk = ll->second.begin();
    set<Expr> old_v = *(kk);
    while (++kk != ll->second.end()) {
      set<Expr> new_v = *(kk);
      for (set<Expr>::iterator ii = old_v.begin(); ii != old_v.end(); ii++) 
	for (set<Expr>::iterator jj = new_v.begin(); jj != new_v.end(); jj++) {
	  ev.push_back(this_dp->mk_lt(*ii, *jj, true));
	}
      old_v = new_v;
    }
  }

  // build expr 
  // for exampe : v+ x + y + z
  // v < x< y <z  
  int index = 0;
  Expr args[ev.size()];
  for (vector<Expr>::iterator ii = ev.begin(); ii != ev.end(); ii++) {
    args[index++] = *ii;
  }
  if (index > 0)
    constraints.push_back(this_dp->mk_and(args, index));
}


Expr AnalysisPF::complimentTransForSharedVar(stack<Expr>& path_cond, set<Expr>& extraTrans) {
  int size = extraTrans.size();
  if (path_cond.size() == 0 || size == 0) {
    return t_expr;
  } 

  Expr p = module_SSA->sum_constraints(path_cond);

  if (size == 1)
    return this_dp->mk_or2(p, *(extraTrans.begin()));
  
  Expr args[size];
  int index = 0;
  for (set<Expr>::iterator ii = extraTrans.begin(); ii != extraTrans.end(); ii++) {
    args[index++] = *ii;
  }
  Expr trans = this_dp->mk_and(args, index);

  return this_dp->mk_or2(p, trans);
}


void  AnalysisPF::reset_sharedVarOrder(ModelPF::VarOrderMap& order) 
{
  makeSharedVarOrder(order, module_SSA->constraints);

  for (ModelPF::VarOrderMap::iterator ii = order.begin(); ii != order.end(); ii++) {
    set<Expr> last_item = ii->second.back();
    ii->second.clear();
    ii->second.push_back(last_item);
  }
}

void AnalysisPF::union_sharedVarOrder(ModelPF::VarOrderMap &l_order, ModelPF::VarOrderMap &r_order, 
				      ModelPF::VarOrderMap &order)
{
  cout << "l_order: " << endl;
  print_order(l_order, this_dp);
  cout << "r_order: " << endl;
  print_order(r_order, this_dp);

  makeSharedVarOrder(l_order, module_SSA->constraints);
  makeSharedVarOrder(r_order, module_SSA->constraints);

  for (ModelPF::VarOrderMap::iterator ii = l_order.begin(); ii != l_order.end(); ii++) {
    string v = ii->first;
    set<Expr> l_last_item = ii->second.back();
    ModelPF::VarOrderMap::iterator jj = r_order.find(v);
    list<set<Expr> > ls;

    if (jj != r_order.end()) {   // in both maps
      set<Expr> r_last_item = jj->second.back();
      std::set<Expr> s = l_last_item;
      for (set<Expr>::iterator ll = r_last_item.begin(); ll != r_last_item.end(); ll++) 
	s.insert(*ll);
      ls.push_back(s);
    }
    else {
      ls.push_back(l_last_item);
    }
    order[v] = ls;
  }

  for (ModelPF::VarOrderMap::iterator ii = r_order.begin(); ii != r_order.end(); ii++) {
    string v = ii->first;
    ModelPF::VarOrderMap::iterator jj = l_order.find(v);
    if (jj == l_order.end()) {
      list<set<Expr> > ls;
      set<Expr> r_last_item = ii->second.back();
      ls.push_back(r_last_item);
      order[v] = ls;
    }
  }

  cout << "order: " << endl;
  print_order(order, this_dp);
}


Expr AnalysisPF::mk_branch_constraint(list<Expr>& exprs) {
  makeSharedVarOrder(module_SSA->sharedVarOrder, module_SSA->constraints);

  exprs.splice(exprs.end(), module_SSA->ps_constraints);  // consider path sensitive constraints too
 
  Expr result = module_SSA->sum_constraints(exprs);
 
  //print_order(module_SSA->sharedVarOrder, this_dp);
  return result;
}

// the last two arguments are for returning some outputs
template<typename T>
Expr AnalysisPF::buildExpr(T& source, list<Expr>& exprs, T& next) {

  FilteredCFGNode<IsDFAFilter> current = source;

  // DFA on that function
  vector< FilteredCFGNode<IsDFAFilter> > worklist;
  worklist.push_back(source);

  while (!worklist.empty()) {
    current = worklist.front();
   // black list region detection
    worklist.erase(worklist.begin());

     //edited by Hongyi
     // For OpenMP supports, we need to pick up the SgOmp* statement as interested from  ompcheckerinterface.cpp.h
     //since MPI definitions are funcitoncall in AST, I am not sure why it is segmentation fault when I add openmp directives with MPI. Now I am debugging MPI first.  
    if (is_interested(current))
      cout << "\n -- visited -- : " << current.toStringForDebugging() << endl;
     // isOmpStatement(current)
      SgNode* sgNodeTmp = current.getNode();
      //!test barrier for MPI and OpenMP 
    
     // there is a bug for master following nodes get_parent
      if(sgNodeTmp->unparseToString().find("MPI_Barrier") == 0)
        {
            std::cout<<" test functioncall barrier expected barrier string "<<sgNodeTmp->unparseToString()<<std::endl;
        }
     // SgFunctionCallExp* fcallExp = isSgFunctionCallExp(sgNodeTmp);
    //  SgFunctionDeclaration *fn = fcallExp->getAssociatedFunctionDeclaration();
      
    //  SgName fnname = fn->get_name();
    //  std::cout<<" test functioncall barrier expected barrier string "<<fnname.unparseToString()<<std::endl;

      //!MPI fucntin call is SgFunctionCall, then we pick up functionCall in if -else branch. Otherwise, no process would identify which MPI routines belong to him. 
     // each vector is a set for each process in MPI. 
     // we removed omp barrier, omp single, omp master, omp flush, omp critical, omp sections for MPI processes debug easily. 
     //after MPI approach is working, then we add these omp features here and ompchecker.cpp for encoding with different threads. 
    
      
      SgPragmaDeclaration* omp = isSgPragmaDeclaration(sgNodeTmp);
      if(omp)
       {
          SgOmpParallelStatement* tmp = (SgOmpParallelStatement*)omp;
          cout<<"\n this block is parallel for "<<tmp->unparseToString()<<endl;
          if(tmp)
          {
              // encoding plus race check
          }
       }
           
   //  cout<<"\n this block is parallel for "<<tmp->unparseToString()<<endl; 
     
     if(omp)
         {
        //   build CFG insdie...
        // parallel encoding
    //    cout<<" parallel !!! "<<tmp->unparseToString()<<endl;
     //   exit(0);
         }    

    vector< FilteredCFGEdge<IsDFAFilter> > out_edges = current.outEdges();
    int out_size = out_edges.size();

    Expr e = visit(current);
    
    if (e != t_expr) {
      exprs.push_back(e);
      print_expr(e);
    }

    // if-else branch return
    if (out_size == 0) {  // the last node of CFG 
      return mk_branch_constraint(exprs);
    }

    FilteredCFGNode<IsDFAFilter> next_node = (*(out_edges.begin())).target();

    if (next_node.inEdges().size() > 1) { //  the next node is a join node
      next = next_node;
      visitedEdges.insert(out_edges.begin()->id());
      return mk_branch_constraint(exprs);
    }

    else if (out_size == 1) {  // ordinary node
      worklist.push_back(next_node);
    }

    else if (out_size == 2) { // a fork node
      vector<filteredCFGEdgeType>::const_iterator i = out_edges.begin();
      SgExpression* cond = ((*i).source().toNode().outEdges().begin())->conditionBasedOn();
      // SgExpression* cond = (*i).caseLabel();

      // process the condition
      stringstream buf;
      buf << "#p" << cond_count++ << "@" <<  module_SSA->thread_id;
      Expr p = this_dp->mk_var(buf.str(), this_dp->mk_bool_type());
      Expr c_exp = module_SSA->convertExp(cond);
      module_SSA->add_constraint(this_dp->mk_eq(p, c_exp));

      // for local variables
      map<string, int> old_localVars = module_SSA->localVars; // push the current local variable map

      filteredCFGEdgeType l_edge = *i;
      filteredCFGNodeType l_target = l_edge.target();
      i++;
      filteredCFGEdgeType r_edge = *i;
      filteredCFGNodeType r_target = r_edge.target();

      Expr e1, e2;
      VarNumMap l_localVars, r_localVars; // the local variable map for the branches
      Expr extraTrans;
      filteredCFGNodeType join1(l_target);
      filteredCFGNodeType join2(r_target);

      ModelPF::ExprList old_ps_constraints = module_SSA->ps_constraints;
      module_SSA->ps_constraints.clear();
      reset_sharedVarOrder(module_SSA->sharedVarOrder);  // keep only the last set of schedule IDs
      ModelPF::VarOrderMap r_order = module_SSA->sharedVarOrder; 

      if (l_target.inEdges().size() > 1) { // dummy edge
	e1 = t_expr;
	join1 = l_target;
	visitedEdges.insert(l_edge.id());
      }
      else {
	path_cond.push(p);  // the left branch
	complimentTrans.clear();

	list<Expr> exprs1;
	e1 = buildExpr(l_target, exprs1, join1);
	print_expr(e1);

	// compliment the transitions 
	extraTrans = complimentTransForSharedVar(path_cond, complimentTrans);
	module_SSA->add_constraint(extraTrans);
	complimentTrans.clear();

	path_cond.pop();
      };
      ModelPF::VarOrderMap l_order = module_SSA->sharedVarOrder; 

      module_SSA->ps_constraints.clear();
      // for global variable ordering
      module_SSA->sharedVarOrder = r_order;

      l_localVars = module_SSA->localVars; // the local variable map for the left branch
      module_SSA->localVars = old_localVars;

      if (r_target.inEdges().size() > 1) { // dummy edge
	e2 = t_expr;
	join2 = r_target;
	visitedEdges.insert(r_edge.id());
      }
      else {
	path_cond.push(this_dp->mk_not(p));                 // the right branch
	
	list<Expr> exprs2;
	e2 = buildExpr(r_target, exprs2, join2);
	print_expr(e2);
	
	// compliment the transitions 
	extraTrans = complimentTransForSharedVar(path_cond, complimentTrans);
	module_SSA->add_constraint(extraTrans);
	complimentTrans.clear();	
      };
      r_order = module_SSA->sharedVarOrder; 
      
      // for local variables
      r_localVars = module_SSA->localVars;      // the local variable map after the right branch is done

      if (DEBUG_MODE) {
	cout << "\nMerge: - " << join1.getNode()->unparseToString() << endl;
	cout <<   "Merge: - " << join2.getNode()->unparseToString() << endl;
      }
      // edited by Hongyi 
    
      // ROSE_ASSERT(join1.toNode().getNode()->unparseToString() == join2.toNode().getNode()->unparseToString());
     //commented by Hongyi
  
  
       // for shared variable ordering
      union_sharedVarOrder(l_order, r_order, module_SSA->sharedVarOrder);

      // solve the discrepancy of SSA indexes between two incoming branches
      pair<Expr,Expr> merge_e = merge_branches(p, l_localVars, r_localVars, module_SSA->localVars);
      // combine the transitions
      Expr combine = this_dp->mk_ite(p, this_dp->mk_and2(e1, merge_e.first), 
				        this_dp->mk_and2(e2, merge_e.second));
      if (combine != t_expr)
	exprs.push_back(combine);
      print_expr(combine);

      module_SSA->ps_constraints = old_ps_constraints;

      // check whether the join node has more than two incoming edges
      vector< FilteredCFGEdge<IsDFAFilter> > in_edges = join1.inEdges();
      if (in_edges.size() > 1) {
	cout << "in edges = " << in_edges.size() << endl;
	bool b = false;
	for (vector< FilteredCFGEdge<IsDFAFilter> >::const_iterator i = in_edges.begin(); i != in_edges.end(); i++) {
	  // cout << "Edge Id = " << i->id() << endl;
	  if (visitedEdges.find(i->id()) == visitedEdges.end())   // still has unvisited edge 
	    {
	      b = true;
	      break;
	    }
	}

	next = join1;
	if (!b) {
	  cout << "adding " << join1.getNode()->unparseToString() << endl;
	  worklist.push_back(join1);
	}
	else {
	  return mk_branch_constraint(exprs);
	}
      }

    }

    else {
      /*
      for (vector< FilteredCFGEdge<IsDFAFilter> >::const_iterator i =
	     out_edges.begin(); i != out_edges.end(); i++) {
	FilteredCFGEdge<IsDFAFilter> filterEdge = *i;
	
	// SgExpression* cond = filterEdge.caseLabel(); //conditionBasedOn();

	if (filterEdge.condition () != eckUnconditional) {
	  SgExpression* cond = (filterEdge.source().toNode().outEdges().begin())->conditionBasedOn();
	  cout << "Cond: " << cond->unparseToString() << endl;
	}
	
	FilteredCFGNode<IsDFAFilter> filterNode = filterEdge.target();
	if (find(worklist.begin(), worklist.end(), filterNode)
	    == worklist.end()) {
	  worklist.push_back(filterNode);
	}
      }
      */
    }
  }
}

//! encoding for each expression in yices
template<typename T>
Expr AnalysisPF::visit(T cfgNode) {
  SgNode* sgNode = cfgNode.getNode();
	
  if (is_interested(cfgNode)) 
   {
     cout <<"expr encoding format: "<<formatExp(sgNode) << endl;
   } 
 //    cout << sgNode->unparseToString() << endl;

  // return true;

  // SgNode* sgNodeBefore = getCFGPredNode(cfgNode);
  ROSE_ASSERT(sgNode);

  bool isDefinition = false;
  SgInitializedName* initName = NULL;
  
  if (isSgUnaryOp(sgNode)) {
    SgUnaryOp* unary = isSgUnaryOp(sgNode);
    SgExpression* l_expr = unary->get_operand();
    SgVarRefExp* varRefExp = NULL;
    
    if (isSgVarRefExp(l_expr)) {
      // if left side is a varrefexp
      varRefExp = isSgVarRefExp(l_expr);
      // initName = varRefExp->get_symbol()->get_declaration();

      Expr lhs = t_expr;
      Expr trans = module_SSA->buildTrans(sgNode, varRefExp, lhs);
      if (lhs != t_expr) 
	complimentTrans.insert(lhs);

      return trans;
    } 
    else {
      if (DEBUG_MODE)
	cout << " **********  UNARY or NonLinear OP : CANT HANDLE YET: "
	     << l_expr->class_name() << endl;
    }
  }
  
  else if (isSgBinaryOp(sgNode)) {
    SgBinaryOp* binary = isSgBinaryOp(sgNode);
    SgExpression* l_expr = binary->get_lhs_operand();
    SgVarRefExp* varRefExp = isSgVarRefExp(l_expr);
    SgPntrArrRefExp* varPntrRefExp = isSgPntrArrRefExp(l_expr);

    if (varRefExp || varPntrRefExp) {
      // if left side is a varrefexp
      if (varRefExp) {
	ROSE_ASSERT(varRefExp->get_symbol());
	initName = varRefExp->get_symbol()->get_declaration();
      }
      else if (varPntrRefExp) {
	while (isSgPntrArrRefExp(varPntrRefExp->get_lhs_operand()))
	  varPntrRefExp = isSgPntrArrRefExp(varPntrRefExp->get_lhs_operand());
	SgVarRefExp* varRefExpL = isSgVarRefExp(varPntrRefExp->get_lhs_operand());
	if (varRefExpL == NULL) {
	  Rose_STL_Container<SgNode*> vars = NodeQuery::querySubTree(varPntrRefExp, V_SgVarRefExp);
	  if (vars.size()>1) {
	    cerr << " There is more than one VarRefExp in this PntrArrRefExp. " << endl;
	    //	    ROSE_ASSERT(varRefExpL);
	    varRefExpL = isSgVarRefExp(*vars.begin());
	  } else if (vars.size()==1)
	    varRefExpL = isSgVarRefExp(*vars.begin());
	  if (varRefExpL==NULL) {
	    cerr << " TYPE of LHS : " << varPntrRefExp->get_lhs_operand()->class_name() << endl;
	    ROSE_ASSERT(varRefExpL);
	  }
	}
	/*
	SgArrowExp* arrow = isSgArrowExp(varPntrRefExp->get_lhs_operand());
	if (arrow) {
	  cerr << " TYPE of LHS OF ARROW: left: " << arrow->get_lhs_operand()->class_name() << endl;
	  cerr << " TYPE of LHS OF ARROW: right: " << arrow->get_rhs_operand()->class_name() << endl;
	  varRefExpL = isSgVarRefExp( arrow->get_rhs_operand());
	  if (varRefExpL==NULL)
	    cerr << " TYPE of LHS IN ARROW: " << arrow->get_rhs_operand()->class_name() << endl;
	  ROSE_ASSERT(varRefExpL);
	}
	if (varRefExpL) {
	  ROSE_ASSERT(varRefExpL);
	  ROSE_ASSERT(varRefExpL->get_symbol());
	} else if (arrow) {
	} else {
	  cerr << " TYPE of LHS : " << varPntrRefExp->get_lhs_operand()->class_name() << endl;
	  ROSE_ASSERT(varRefExpL);
	}
	*/
	
	initName = varRefExpL->get_symbol()->get_declaration();
      }
      
      Expr lhs = t_expr;
      Expr trans = module_SSA->buildTrans(sgNode, l_expr, lhs);
      if (lhs != t_expr) 
	complimentTrans.insert(lhs);
      return trans;
    }
    /*
    else {
      if (DEBUG_MODE)
	cout << " **********  BINARY OP : CANT HANDLE YET: " << l_expr->class_name() << endl;
    }
    */
  }

  else if (isSgAssignInitializer(sgNode)) {
    SgAssignInitializer* initializer = isSgAssignInitializer(sgNode);
    SgNode* l_expr = isSgAssignInitializer(sgNode)->get_parent();
    ROSE_ASSERT(l_expr);
    if (isSgInitializedName(l_expr)) {
      initName = isSgInitializedName(l_expr);

      Expr lhs = t_expr;
      Expr trans = module_SSA->buildTrans(sgNode, initName, lhs);
      if (lhs != t_expr) 
	complimentTrans.insert(lhs);

      return trans;

    } else if (isSgExprListExp(l_expr)) {
      SgExprListExp* listexp = isSgExprListExp(l_expr);
      Rose_STL_Container<SgNode *> exprs = listexp->get_traversalSuccessorContainer();
      // exprs.size()
      cout << " unparse : " << l_expr->unparseToCompleteString() << endl;
      return t_expr;
    } else {
      cout << " Analysis:: unhandled condition - AssignInitializer ... "
	   << "  AssignInitializer - l_expr : " << l_expr->class_name() << "  file : " 
	   << l_expr->get_file_info()->get_filenameString() << "  line : " << l_expr->get_file_info()->get_line() << endl;
      cout << " unparse : " << l_expr->unparseToCompleteString() << endl;
      return t_expr;

    }
  }

  else if (isSgFunctionRefExp(sgNode)) {
    module_SSA->convertExp(sgNode);
  }

  /*  
  else if (isSgInitializedName(sgNode)) {
    initName =isSgInitializedName(sgNode);
    if (DEBUG_MODE)
      cout << " **********  INITNAME. " << initName << endl;
    //isUse=false;
    isDefinition=true;
    if (DEBUG_MODE)
      cout << " **********  INITNAME. "
	   << " def " << isDefinition << endl;
  }
  */

  return t_expr;

}

void AnalysisPF::check(Expr e) {

  this_dp->DP_assert(e);
  // this_dp->dump_context();

  switch(this_dp->check()) {
  case l_true: {
    printf("satisfiable\n");
    DP_model m = this_dp->get_model();
    this_dp->display_model(m);
    break;  
  }
  case l_false:
    printf("unsatisfiable\n");
    break;
  case l_undef:
    printf("unknown\n");
    break;
  }
}


/******************************************
 * Traversal over all functions
 *****************************************/
Expr  AnalysisPF::start_traversal_of_functions() {
  if (DEBUG_MODE) 
    cout << "START: Traversal over Functions" << endl;

  nrOfNodesVisitedPF = 0;

  // Traverse through each FunctionDefinition
  Rose_STL_Container<SgNode*> functions = NodeQuery::querySubTree(project, V_SgFunctionDefinition); 

  Expr trans = t_expr;

  for (Rose_STL_Container<SgNode*>::const_iterator i = functions.begin(); i != functions.end(); ++i) {
    SgFunctionDefinition* proc = isSgFunctionDefinition(*i);

    bool abortme = false;
    trans = run(proc, abortme);
    nrOfNodesVisitedPF += getNumberOfNodesVisited();
  }
  // delete perfunc;

  if (DEBUG_MODE) 
    cout << "FINISH: Traversal over test  Functions" << endl;
  
  return trans;
}

/******************************************
 * Traversal the function Declaration
 *****************************************/

 // only function call and main function
Expr  AnalysisPF::start_traversal_of_functions(SgFunctionDefinition* proc) {
  if (DEBUG_MODE) 
    cout << "START: Traversal over the  function declarations" << endl;

  nrOfNodesVisitedPF = 0;

  Expr trans = t_expr;
  bool abortme = false;
  trans = run(proc, abortme);
  nrOfNodesVisitedPF += getNumberOfNodesVisited();
  // delete perfunc;
  if (DEBUG_MODE) 
    cout << "FINISH: Traversal over the function decalaration, not only main" << endl;
  return trans;
}

/**********************************************************
 *  Build CFG for each function and do analysis
 *********************************************************/

Expr AnalysisPF::run(SgFunctionDefinition* funcDecl, bool& abortme) {

  string funcName = funcDecl->get_declaration()->get_name().str();
  SgFunctionDeclaration* funcDeclaration = funcDecl->get_declaration();
 
 //! edited by Hongyi
 //all the functions shoud be analyzed
 
  if (!is_check_func(funcDeclaration) && !is_global_func(funcDeclaration)) {
   //   return t_expr;
  }
  ROSE_ASSERT(funcDecl);
 //end 
  if (DEBUG_MODE)
    cout << " Found function: " << funcName << endl;
   cout << " Found function and begin to check functions " << funcName << endl;

  // build CFG, currently, we can get CFG inside SgOmp*

  // add this node to worklist and work through the outgoing edges
  FilteredCFGNode<IsDFAFilter> source = 
    FilteredCFGNode<IsDFAFilter> (funcDecl->cfgForBeginning());

  //! this is for debug mode
  
  if (DEBUG_MODE) {
    std::ofstream f("debugCFG.dot");
    cfgToDot(f, string("cfg"), source);
    f.close();
  }

  FilteredCFGNode<IsDFAFilter> next = source;
  visitedEdges.clear();
  list<Expr> exprs;
  Expr e = buildExpr(source, exprs, next);
   
  /*
  cout << endl << endl;
  print_expr(e);
  cout << "Local constraints:" << endl;
  module_SSA->print_constraints(module_SSA->constraints);
  */

  // check(e);

  if (DEBUG_MODE) {
    cout << " Exiting function " << funcName << endl;
  }
  return e;
}

/******************************************
 * Interface
 *****************************************/
/*
int main( int argc, char * argv[]) {
  SgProject* project = frontend(argc, argv);
  std::cout << ">>>> start analysis ... " << endl;

  DP dp;
  AnalysisPF* analysis = new AnalysisPF(&dp,true);
  analysis->project = project;
  analysis->start_traversal_of_functions();

  delete analysis;

  return 0;
}
*/
