
#ifndef __ANALYSIS_PF__
#define __ANALYSIS_PF__

#include "rose.h"
#include <cfgToDot.h>
#include <list>
#include "filteredCFG.h"
#include "DFAFilter.h"
#include <string>
#include <map>

#include "modelPF.h"

using namespace std;

class AnalysisPF  {
 private:
    bool DEBUG_MODE;
    int nrOfNodesVisitedPF;

    Expr t_expr;
    set<Expr> complimentTrans;    // transitions for the lhs variables in a block 

    int cond_count;
    stack<Expr> path_cond;
    list<Expr> constraints;

    // for testing
    void print_expr(Expr e) {
      module_SSA->dp->pp_expr(e);
      cout << endl;
    };

 public:
    ModelPF* module_SSA;
    SgProject* project;

    AnalysisPF(DP* dp_in, bool debug) {
      DEBUG_MODE=debug;
      module_SSA = new ModelPF(dp_in, debug);
      t_expr = dp_in->mk_true();
      cond_count = 0;
    };
    virtual ~AnalysisPF(){
      delete module_SSA;
    };

    Expr start_traversal_of_functions();
    Expr start_traversal_of_functions(SgFunctionDefinition* proc);

    inline int getNumberOfNodesVisited() {
      return nrOfNodesVisitedPF;
    };
    int breakPointForWhile;
    SgNode* breakPointForWhileNode;
    
    Expr run(SgFunctionDefinition* function, bool& abortme);
    typedef FilteredCFGEdge < IsDFAFilter > filteredCFGEdgeType;
    typedef FilteredCFGNode < IsDFAFilter > filteredCFGNodeType;
    typedef map<string, int> VarNumMap;
    set<string> visitedEdges;

    void reset_sharedVarOrder(ModelPF::VarOrderMap& order);
    void union_sharedVarOrder(ModelPF::VarOrderMap &l_order, ModelPF::VarOrderMap &r_order, 
			      ModelPF::VarOrderMap &order);
    void makeSharedVarOrder(ModelPF::VarOrderMap &sharedVarOder, list<Expr>& constraints);
    Expr complimentTransForSharedVar(stack<Expr>& path_cond, set<Expr>& extraTrans);
    pair<Expr,Expr> merge_branches(Expr p, VarNumMap& l_map, VarNumMap& r_map, VarNumMap& result_map);
    Expr mk_branch_constraint(list<Expr>& exprs);
    
    template<typename T> Expr visit(T cfgNode);
    template<typename T> Expr buildExpr(T& cfgNode, list<Expr>& exprs, T& next);

    void check(Expr e);

};

#endif

