#ifndef __MY_DATASTRUCT__
#define __MY_DATASTRUCT__

#include "dp.h"
//! last edited by Hongyi on 11th September 2013.
using namespace std;

struct Variable {
  string name;
  int dim;     // dimension for array element, 0 indicates a non-array variable
  bool is_shared;
  bool gurantee; // since MPI is distributed, so some variables are independent.
  bool assume;   // sections may have races
  // vector<Expr> args;

  bool initialized;
  Type type;

  Variable() {
    dim = 0;
    initialized = false;
    is_shared = false;

  }

  Variable(string name1, int dim_size, bool is_shared1) {
    dim = dim_size;
    initialized = false;
    name = name1;
    is_shared = is_shared1;
  }

  inline bool is_array() {
    return dim > 0;
  }

  Type get_var_type(DP* dp) {
    if (initialized)
      return type;
    else if (dim > 0) {  // an array variable
      Type basic_ty = dp->mk_basic_type();
      Type domain[dim];
      
      for (int i = 0; i < dim; i++) { 
	domain[i] = basic_ty;
      }
      type = dp->mk_function_type(domain, dim, basic_ty);

      if (is_shared) {    // f::nat->(basic->basic->...->basic)
	type = dp->mk_unary_function_type(dp->mk_nat_type(), type);
      }

      initialized = true;
      return type;
    }
    else {
      Type basic_ty = dp->mk_basic_type();
      if  (is_shared) 
	type = dp->mk_unary_function_type(dp->mk_nat_type(), basic_ty);
      else 
	type = basic_ty;
      initialized = true;
      return type;
    }
  }

  Expr get_var_expr(DP* dp) {
    Type tp = get_var_type(dp);
    Expr var = dp->mk_var(name, tp);
    return var;
  }

  /* Operator overloading */

  Variable& operator=(const Variable& rhs) {  // assignment Operator 
    name = rhs.name;
    dim = rhs.dim;
    initialized = rhs.initialized;
    type = rhs.type;  
    is_shared = rhs.is_shared;
  };

  bool operator==(const Variable& other) const {
    return name == other.name;
  }

  bool operator!=(const Variable& other) const {
    return name != other.name;
  }

  bool operator<(const Variable& other) const {
    return name < other.name;
  }

  bool operator<=(const Variable& other) const {
    return name <= other.name;
  }

};

ostream &operator<<(ostream &stream, Variable var);

#endif
