#include "dp.h"

/**********************************************************
   **********************************************************/

Type DP::mk_bool_type() {
  return yices_mk_type(ctx, "bool");
}

Type DP::mk_nat_type() {
  return yices_mk_type(ctx, "nat");
}

Type DP::mk_int_type() {
  return yices_mk_type(ctx, "int");
}

/**********************************************************
 **********************************************************/

Type DP::mk_unary_function_type(Type domain1, Type range) {
  Type domain[1] = {domain1};
  return yices_mk_function_type (ctx, domain, 1, range);
}

Var_decl DP::mk_var_decl(std::string vname, Type ty) {
  char* s = (char*) (&(vname[0])); // .c_str();
  return yices_mk_var_decl(ctx, s, ty);
}

Expr DP::get_var_decl_from_name (std::string name) {
  return yices_get_var_decl_from_name(ctx, (char*) (name.c_str()));
}

Expr DP::get_var_from_name (std::string name) {
  return mk_var_from_decl(get_var_decl_from_name(name));
}

Expr DP::mk_var(std::string vname, Type ty) {
  Var_decl var_decl;
  try {
    var_decl = mk_var_decl(vname, ty); 
  } catch (...)
    {
      var_decl = get_var_decl_from_name (vname); 
    }
  return mk_var_from_decl(var_decl);
}


/**********************************************************
 **********************************************************/



Expr DP::mk_sum2(Expr x,  Expr y, bool use_int)
  {
#ifdef USE_INT
    Expr args[2] = {x,y}; 
    return mk_sum (args, 2);
#else
    if (use_int) { 
      Expr args[2] = {x,y}; 
      return mk_sum (args, 2);
    }
    else 
      return yices_mk_bv_add(ctx, x, y);
#endif
  }

Expr DP::mk_sub2(Expr x,  Expr y, bool use_int)
  {
#ifdef USE_INT
    Expr args[2] = {x,y}; 
    return mk_sub (args, 2);
#else
    if (use_int) { 
      Expr args[2] = {x,y}; 
      return mk_sub (args, 2);
    }
    else {
       return yices_mk_bv_sub(ctx, x, y);
    }
#endif
  }

Expr DP::mk_mul2(Expr x,  Expr y, bool use_int)
  {
#ifdef USE_INT
    Expr args[2] = {x,y}; 
    return mk_mul (args, 2);
#else
    if (use_int) { 
      Expr args[2] = {x,y}; 
      return mk_mul (args, 2);
    }
    else {
      return yices_mk_bv_mul(ctx, x, y);
    }
#endif
  }
