
#ifndef __MY_DP__
#define __MY_DP__

#include <string>
#include "yices_c.h"
#include "yicesl_c.h"

typedef yices_context Context;
typedef yices_type Type;
typedef yices_var_decl Var_decl;
typedef yices_expr Expr; 
typedef yices_model DP_model; 

#define NUM_OF_BITS 4
// #define USE_INT

// new undocumented function/hack
extern "C" {
  extern yicesl_context yices_get_lite_context(yices_context ctx);
}

//store yices format for SMT-solver after encoding
class DP {

  Context ctx;
  Context hidden;

 public:

  DP() {
    ctx = yices_mk_context();
    hidden = yices_get_lite_context(ctx);
    
    t_expr = mk_true();
    enable_type_checker (true);
  };

  ~DP() {
    yices_del_context(ctx);
  };

  Expr t_expr;

  inline int read_cmd(std::string cmd) {
    yicesl_read (hidden, (char*) (cmd.c_str()));
  }

  /**********************************************************
   **********************************************************/

  inline void DP_assert(Expr expr) {
    yices_assert (ctx, expr);
  }

  inline void dump_context() {
    yices_dump_context(ctx);
  }
  
  inline lbool check() {
    return yices_check (ctx);
  }
  
  inline DP_model get_model() {
    return  yices_get_model(ctx);
  } 
  
  inline void display_model(DP_model model) {
    return  yices_display_model(model);
  }
  
  inline lbool get_value(DP_model m, Var_decl v) {
    return yices_get_value (m, v);
  }
  
  inline void pp_expr (Expr e) {
    return yices_pp_expr (e);
  }
  
  inline void enable_type_checker (int flag) {
    return yices_enable_type_checker (flag);
  }
  
  /**********************************************************
   **********************************************************/

  inline Expr mk_fresh_bool_var() {
    return yices_mk_fresh_bool_var(ctx);
  }

  inline Type mk_type(char* tname) {
    return yices_mk_type(ctx, tname);
  }

  // show-statistic 
  //edited by Hongyi 

  inline long long get_cost(yices_model m)
   {
     return yices_get_cost(m);
   } 

  Type mk_bool_type();
  Type mk_nat_type();
  Type mk_int_type();

  inline Type mk_bitvector_type (unsigned size) {
    return yices_mk_bitvector_type (ctx, size);
  }

  inline Type mk_basic_type() {
#ifdef USE_INT
    return mk_int_type();
#else
    return mk_bitvector_type(NUM_OF_BITS);
#endif
  }

  inline Type mk_function_type(Type domain[], int domain_size, Type range) {
    return yices_mk_function_type (ctx, domain, domain_size, range);
  }

  Type mk_unary_function_type(Type domain1, Type range);

// yices_mk_function_type

  Var_decl mk_var_decl(std::string vname, Type ty);
 
/*
  inline Var_decl mk_var_decl(char* vname, Type ty) {
    return yices_mk_var_decl(ctx, vname, ty);
  }
*/

  inline Expr mk_var_from_decl(Var_decl decl) {
    return yices_mk_var_from_decl(ctx, decl);
  }

  Expr get_var_decl_from_name (std::string name);
  Expr get_var_from_name (std::string name);
  Expr mk_var(std::string vname, Type ty); 

  inline Expr mk_int_num(int n) { 
    return yices_mk_num (ctx, n);
  }

  inline Expr mk_num(int n) {
    #ifdef USE_INT    
       return yices_mk_num (ctx, n);
    #else
       return yices_mk_bv_constant (ctx, NUM_OF_BITS, n);
    #endif
  }
  
  /**********************************************************
   **********************************************************/

  inline Expr mk_true() {
    return yices_mk_true (ctx);
  }

  inline Expr mk_false() {
    return yices_mk_false (ctx);
  }

  inline Expr mk_not(Expr x) {
    return yices_mk_not (ctx, x);
  }

  Expr mk_and2(Expr x, Expr y) {
    if (x == t_expr) return y;
    else if (y == t_expr) return x;

    Expr args[2] = {x,y}; 
    return yices_mk_and (ctx, args, 2);
  }

  Expr mk_and(Expr args[], int n) {
    if (n == 0) return t_expr; 
    else return n == 1 ? args[0] : yices_mk_and (ctx, args, n);
  }

  inline Expr mk_or2(Expr x, Expr y) {
    Expr args[2] = {x,y}; 
    return yices_mk_or (ctx, args, 2);
  }

  Expr mk_or(Expr args[], int n) {
    if (n == 0) return t_expr; 
    return n == 1 ? args[0] : yices_mk_or (ctx, args, n);
  }

  Expr mk_imp(Expr x, Expr y) {
    if (x == t_expr) 
      return y;
    else
      return mk_or2(mk_not(x), y);
  }

  inline Expr mk_ite(Expr c, Expr t, Expr e) {
    return yices_mk_ite (ctx, c, t, e);
  }

  /**********************************************************
   **********************************************************/

  inline Expr mk_app(Expr f, Expr args[], int n) {
    return yices_mk_app(ctx, f, args, n);
  }

  inline Expr mk_unary_app(Expr f, Expr arg) {
    Expr args[1] = {arg};
    return yices_mk_app(ctx, f, args, 1);
  }

  inline Expr mk_eq(Expr x, Expr y)
  {
    return yices_mk_eq(ctx, x, y);
  }
  
  inline Expr mk_diseq(Expr x, Expr y)
  {
    return yices_mk_diseq(ctx, x, y);
  }

  /**********************************************************
   **********************************************************/
  
  inline Expr mk_gt(Expr x, Expr y, bool use_int = false)
  {
#ifdef USE_INT
    return yices_mk_gt(ctx, x, y);
#else
    return use_int ? yices_mk_gt(ctx, x, y) : yices_mk_bv_gt(ctx, x, y);
#endif
  }
  
  inline Expr mk_ge(Expr x, Expr y, bool use_int = false)
  {
#ifdef USE_INT
    return yices_mk_ge(ctx, x, y);
#else
    return use_int ? yices_mk_ge(ctx, x, y) : yices_mk_bv_ge(ctx, x, y);
#endif
  }
  
  inline Expr mk_lt(Expr x, Expr y, bool use_int = false)
  {
#ifdef USE_INT
    return yices_mk_lt(ctx, x, y);
#else
    return use_int ? yices_mk_lt(ctx, x, y) : yices_mk_bv_lt(ctx, x, y);
#endif
  }
  
  inline Expr mk_le(Expr x, Expr y, bool use_int = false)
  {
#ifdef USE_INT
    return yices_mk_le(ctx, x, y);
#else
    return  use_int ? yices_mk_le(ctx, x, y) : yices_mk_bv_le(ctx, x, y);
#endif
  }
  
  /**********************************************************
   **********************************************************/

  inline Expr mk_sum(Expr args[],  unsigned n)
  {
    return yices_mk_sum (ctx, args, n);
  }
  
  inline Expr mk_sub(Expr args[],  unsigned n)
  {
    return yices_mk_sub (ctx, args, n);
  }

  inline Expr mk_mul(Expr args[],  unsigned n)
  {
    return yices_mk_mul (ctx, args, n);
  }

  Expr mk_sum2(Expr x,  Expr y, bool use_int = false);
  Expr mk_sub2(Expr x,  Expr y, bool use_int = false);
  Expr mk_mul2(Expr x,  Expr y, bool use_int = false);

  inline Expr mk_bv_and(Expr x, Expr y) {
    return yices_mk_bv_and (ctx, x, y);
  }

  inline Expr mk_bv_or(Expr x, Expr y) {
    return yices_mk_bv_or (ctx, x, y);
  }

  inline Expr mk_bv_xor(Expr x, Expr y) {
    return yices_mk_bv_xor (ctx, x, y);
  }

  inline Expr mk_bv_not(Expr x) {
    return yices_mk_bv_not (ctx, x);
  }

  inline Expr mk_bv_shift_left(Expr e,  unsigned n)
  {
    return yices_mk_bv_shift_left0 (ctx, e, n);
  }
  
  inline Expr mk_bv_shift_right(Expr e,  unsigned n)
  {
    return yices_mk_bv_shift_right0 (ctx, e, n);
  }

  /**********************************************************
   **********************************************************/

  inline Expr mk_bv_constant_from_array (unsigned size, int bv[])
  {
    return yices_mk_bv_constant_from_array (ctx, size, bv);
  }

  inline Expr num_from_string (char *n) 
  {
    return yices_mk_num_from_string (ctx, n);
  }

};

#endif
