//! edited by Hongyi
// this is to work on array dimension parse and ssa updates.
#include "modelPF.h"
#include "util.h"

using namespace std;

//no partial order recution
bool Config::use_POR = false;

//! 
SgInitializedName* getRefVar(SgNode* refExp) {
  if (isSgPntrArrRefExp(refExp)) {
    SgPntrArrRefExp* varPntrRefExp = isSgPntrArrRefExp(refExp);

    while (isSgPntrArrRefExp(varPntrRefExp->get_lhs_operand())) {
      varPntrRefExp = isSgPntrArrRefExp(varPntrRefExp->get_lhs_operand());
    }
    // currently for mesesage race, we care about send/reciever with scheudle id   
    // matrix name 
    SgVarRefExp* exp = isSgVarRefExp(varPntrRefExp->get_lhs_operand());
    ROSE_ASSERT(exp != NULL);
    return exp->get_symbol()->get_declaration();
  }
  else if (isSgVarRefExp(refExp)) {
    SgVarRefExp* exp = isSgVarRefExp(refExp);
    ROSE_ASSERT(exp != NULL);
    return exp->get_symbol()->get_declaration();
  }
  else if (isSgInitializedName(refExp)) {
    return isSgInitializedName(refExp);
  }
  return NULL;
}

void ModelPF::record_variable(SgNode* refExp) 
{
  SgInitializedName* initName = getRefVar(refExp);
  string name = initName->get_name().str();  
  if (sharedVars.find(name) != sharedVars.end() || localVars.find(name) != localVars.end()) {
    return;
  }

  int dim_size = 0;
  bool is_shared = false;
  
  if (isSgPntrArrRefExp(refExp)) {
    SgPntrArrRefExp* varPntrRefExp = isSgPntrArrRefExp(refExp);

    // cout << "calculate the dim size " << endl;
    dim_size = 1;
    while (isSgPntrArrRefExp(varPntrRefExp->get_lhs_operand())) {
      varPntrRefExp = isSgPntrArrRefExp(varPntrRefExp->get_lhs_operand());
      dim_size++;
    }
    // cout << "dim_size = " << dim_size << endl;
  }
   // this part is conducted by shared, private clause list and local variables definitons
   // cout << "name = " << name << endl;
  if (is_shared_var(initName) ) {
    // cout << "Found a shared var " << name << endl;
    is_shared = true;
  }
    
  lookupVars[name] = Variable(name, dim_size, is_shared);
      
  if (is_shared) { 
    sharedVars[name] = 0; // set global variables' SSA to 0
    sharedTerms[name] = new ExprList();

    // initialize the global order list
    set<Expr> empty_set;
    list<set<Expr> > ls;
    ls.push_back(empty_set);
    sharedVarOrder[name] = ls;
  }
  else {
    localVars[name] = 0;
  }
  
  /*
    if (check_race) {
    try {
    dp->mk_var_decl("$" + *ii, n_n_ty);  // declare the varable for race check  
    } catch (...) 
    {}
    }
  */
}

bool ModelPF::isBuiltInExp(SgNode* sgNode, Expr &t) 
{
  if (isSgDotExp(sgNode)) {
    SgDotExp* dotExp = isSgDotExp(sgNode);
    SgExpression* l_expr = dotExp->get_lhs_operand();
    SgExpression* r_expr = dotExp->get_rhs_operand();
   
    if (isSgVarRefExp(l_expr) && isSgVarRefExp(l_expr)) {
      SgInitializedName* l_initName = isSgVarRefExp(l_expr)->get_symbol()->get_declaration();
      SgInitializedName* r_initName = isSgVarRefExp(r_expr)->get_symbol()->get_declaration();
      string l_s = l_initName->get_name().str();
      string r_s = r_initName->get_name().str();

      
      // edited by Hongyi 
      // remvoe thread,  schedule id , or variable x
      //! if you are interested in some variable, so please modify them as buildIn. and return true.
      }
    }

  return false;
}

Expr ModelPF::convertExp(SgNode* sgNode, map<string,Expr>* vmap)
{
  Expr t;
  if (isBuiltInExp(sgNode, t)) {
     // edited by Hongyi 
     // test built in expr 
     // buildIn may verify whether your desired variable are able to be encoded.
     return t;
  }

  if (isSgVarRefExp(sgNode)) {   // variables
    SgVarRefExp* varRefExp = isSgVarRefExp(sgNode);
    // if it is shared
    record_variable(varRefExp);

    SgInitializedName* initName = varRefExp->get_symbol()->get_declaration();
    if (vmap != NULL)
      {
       // find schedule ssa information
      return construct_var(initName, vmap);
      }
    else
      { // update ssa information for encoding
       return obtain_SSA_var(initName);
      }
  }
  
  else if (isSgPntrArrRefExp(sgNode)) {

    record_variable(isSgPntrArrRefExp(sgNode));

    stack<Expr> args_stack;
    SgNode* expression = sgNode;
    do {
      SgPntrArrRefExp* arrayReference = isSgPntrArrRefExp(expression);
      SgExpression* indexExpression = arrayReference->get_rhs_operand();
      args_stack.push(convertExp(indexExpression, vmap));
      expression = arrayReference->get_lhs_operand();
    }
    while (isSgPntrArrRefExp(expression));

    SgVarRefExp* arrayVarRef = isSgVarRefExp(expression);
    ROSE_ASSERT(arrayVarRef != NULL);
    SgInitializedName* initName = arrayVarRef->get_symbol()->get_declaration();

    Expr arrayVar;
    if (vmap != NULL)
      arrayVar = construct_var(initName, vmap);
    else
      arrayVar = obtain_SSA_var(initName);

    Expr args[args_stack.size()];
    int index = 0;
    while (!args_stack.empty()) {
      args[index++] = args_stack.top();
      args_stack.pop();
    }

    return dp->mk_app(arrayVar, args, index);
  }

  else if (isSgValueExp(sgNode)) { // constants
    SgValueExp* varRef = isSgValueExp(sgNode);
    // SgName name = varRef->get_symbol()->get_name();
    // return varRef->class_name();
    // stringstream out;
    
    switch(varRef->variantT()) {
    case V_SgIntVal:
    case V_SgUnsignedIntVal:
      // out << isSgIntVal(varRef)->get_value();
      return dp->mk_num(get_num_value(sgNode));
    case V_SgBoolValExp:
      return get_num_value(sgNode) == 0 ? dp->mk_false() : dp->mk_true();
      /*
    case V_SgFloatVal:
      out << isSgFloatVal(varRef)->get_value();
      return out.str();
    case V_SgDoubleVal:
      out << isSgDoubleVal(varRef)->get_value();
      return out.str();
      */
    default:
      return dp->mk_num(0);
    }
  }

  else if (isSgUnaryOp(sgNode)) {
    SgUnaryOp* unary = isSgUnaryOp(sgNode);
    SgExpression* l_expr = unary->get_operand();
    switch (unary->variantT()) {
    case V_SgPlusPlusOp:
      return dp->mk_sum2(convertExp(l_expr, vmap), dp->mk_num(1));
    case V_SgMinusMinusOp: 
      return dp->mk_sub2(convertExp(l_expr, vmap), dp->mk_num(1));
    case V_SgNotOp: 
      return dp->mk_not(convertExp(l_expr, vmap));
 
    default:
      return dp->mk_num(0);
    }
  }

  else if (isSgBinaryOp(sgNode)) {
    SgBinaryOp* binary = isSgBinaryOp(sgNode);
    SgExpression* l_expr = binary->get_lhs_operand();
    SgExpression* r_expr = binary->get_rhs_operand();

    switch(binary->variantT()) {
    case V_SgAddOp:
      return dp->mk_sum2(convertExp(l_expr, vmap), convertExp(r_expr, vmap));
    case V_SgSubtractOp:
      return dp->mk_sub2(convertExp(l_expr, vmap), convertExp(r_expr, vmap));
    case V_SgMultiplyOp:
      return dp->mk_mul2(convertExp(l_expr, vmap), convertExp(r_expr, vmap));

    case V_SgLshiftOp:
      return dp->mk_bv_shift_left(convertExp(l_expr, vmap), get_num_value(r_expr));
    case V_SgRshiftOp:
      return dp->mk_bv_shift_right(convertExp(l_expr, vmap), get_num_value(r_expr));
    case V_SgBitAndOp:
      return dp->mk_bv_and(convertExp(l_expr, vmap), convertExp(r_expr, vmap));
    case V_SgBitOrOp:
      return dp->mk_bv_or(convertExp(l_expr, vmap), convertExp(r_expr, vmap));
    case V_SgBitXorOp:
      return dp->mk_bv_xor(convertExp(l_expr, vmap), convertExp(r_expr, vmap));

    case V_SgEqualityOp:
      return dp->mk_eq(convertExp(l_expr, vmap), convertExp(r_expr, vmap));
    case V_SgNotEqualOp:
      return dp->mk_diseq(convertExp(l_expr, vmap), convertExp(r_expr, vmap));
    case V_SgLessThanOp:
      return dp->mk_lt(convertExp(l_expr, vmap), convertExp(r_expr, vmap));
    case V_SgGreaterThanOp:
      return dp->mk_gt(convertExp(l_expr, vmap), convertExp(r_expr, vmap));
    case V_SgLessOrEqualOp:
      return dp->mk_le(convertExp(l_expr, vmap), convertExp(r_expr, vmap));
    case V_SgGreaterOrEqualOp:
      return dp->mk_ge(convertExp(l_expr, vmap), convertExp(r_expr, vmap));

    case V_SgAndOp:
      return dp->mk_and2(convertExp(l_expr, vmap), convertExp(r_expr, vmap));
    case V_SgOrOp:
      return dp->mk_or2(convertExp(l_expr, vmap), convertExp(r_expr, vmap));


    // other expressions
    case V_SgDotExp:
      return dp->mk_num(0);

    default:
      return dp->mk_true();
    }
  }

  else if (isSgFunctionRefExp(sgNode)) {
    SgFunctionRefExp *functionRefExp = isSgFunctionRefExp(sgNode);
    SgFunctionSymbol *symbol = functionRefExp->get_symbol();
    //edited by Hongyi 
    // in order to support MPI 
    //For Both MPI and OpenMP barriers, we comapred with name instead of node class 
    // __syncthreads for threads and processes
    //edited by Hongyi 
    // For MPI barrier, 
    //! it is expected to utilize 
    
  //   functionRefExp->unparseToString().find("MPI_Barrier") 
   
     if ((strcmp(symbol->get_name().str(), "MPI_Barrier(MPI_COMM_WORLD)") == 0) ||(strcmp(symbol->get_name().str(), "barrier") == 0) )
      {
        std::cout<<"MPI Barrier is for bar var variable created" <<std::endl;
        //exit(0);
	if (DEBUG_MODE) 
	  cout << "Found the call of syncthreads and syncprocesses" << endl;

	Var_decl var_decl;

	// enable the ordering constraint
	for (ModelPF::VarOrderMap::iterator ii = sharedVarOrder.begin(); ii != sharedVarOrder.end(); ii++) {
	  string vname = ii->first;
	  stringstream out;
	  out << "bar" << barrier_count << "." << vname;
	  Expr bar_var = dp->mk_var(out.str(), dp->mk_nat_type());
	  barriers[vname].push_back(bar_var);
	  
	  if (Config::use_POR)  {  // use_POR
	    sharedTerms[vname]->push_back(bar_var);
	  }

	  set<Expr> s;
	  s.insert(bar_var);
	  ii->second.push_back(s);

	  // the value of all shared variables are unchanged, i.e. the barrier is blocked.
	  Variable var = lookupVars[ii->first];
	  Expr trans = mk_unchanged_trans(var.get_var_expr(dp), bar_var);
	  add_constraint(trans);
	}

	barrier_count++;

 //edited by Hongyi to chece race 
       //if (check_race) {
  	  // unset the rw value
  	 // Expr args[1] = {var};
	
//   	  Expr rw_var = dp->mk_app(rw, args, 1);
  //	  Expr rw_id_exp = dp->mk_eq(rw_var, NU_num); 
 //	  add_constraint(rw_id_exp);
 //	}
        // end of check race
      }
  }
  /*
  case V_SgFunctionCallExp:
  {
    SgFunctionCallExp* fcallExp = isSgFunctionCallExp(sgNode);
    SgExprListExp* exprList = fcallExp->get_args();
    SgExpressionPtrList& list = exprList->get_expressions();

    SgExpressionPtrList::iterator i;
    string s = "";
    for (i = list.begin(); i != list.end(); ++i) {
      SgExpression* expr = *i;
      s = s + formatExp(expr);
    }
    return fcallExp + "(" + s + ")";
  }
  */
  else {
    cout << "Encounter " << sgNode->unparseToString() << endl;
    return dp->mk_num(0);
  }
}

Expr ModelPF::makeGlobalVarIntactTrans(string v) {
  stringstream out;
  out << IN_PROC(v, sharedVars[v], thread_id);  // schedule id
  string name = out.str();

  cout << "global variable name = " << name << endl;
	
  Expr schedule_id = dp->mk_var(name, dp->mk_nat_type());
  Expr var = dp->get_var_from_name(v);
  Expr new_var = dp->mk_unary_app(var, schedule_id);

  Expr previous_id = dp->mk_sub2(schedule_id, dp->mk_int_num(1), true);
  Expr old_var = dp->mk_unary_app(var, previous_id);
	  
  return dp->mk_eq(new_var, old_var);
}


Expr ModelPF::make_update_array_trans(SgNode* refExp, Expr value)
{
  record_variable(refExp);

  if (isSgInitializedName(refExp)) { 
    Expr var = redefine_var(isSgInitializedName(refExp));
    Expr trans = dp->mk_eq(var, value);
    return trans;
  }
  else if (isSgVarRefExp(refExp)) { 
    SgVarRefExp* varRefExp = isSgVarRefExp(refExp);
    Expr var = redefine_var(varRefExp->get_symbol()->get_declaration());
    Expr trans = dp->mk_eq(var, value);
    return trans;
  }

  else if (isSgPntrArrRefExp(refExp)) {
    SgPntrArrRefExp* varPntrRefExp = isSgPntrArrRefExp(refExp);
    cout << "array update " << endl;

    stack<Expr> args_stack;
    SgNode* expression = refExp;
    do {
      SgPntrArrRefExp* arrayReference = isSgPntrArrRefExp(expression);
      SgExpression* indexExpression = arrayReference->get_rhs_operand();
      args_stack.push(convertExp(indexExpression));
      expression = arrayReference->get_lhs_operand();
    }
    while (isSgPntrArrRefExp(expression));

    SgVarRefExp* arrayVarRef = isSgVarRefExp(expression);
    ROSE_ASSERT(arrayVarRef != NULL);
    SgInitializedName* initName = arrayVarRef->get_symbol()->get_declaration();
    string vname = initName->get_name().str();
    Variable v = lookupVars[vname];

    cout << "id = " << sharedVars[vname] << endl;

    Expr old_arrayVar, arrayVar;
    if (!v.is_shared) {
      old_arrayVar = obtain_SSA_var(initName, false);
      arrayVar = redefine_var(initName);
    }
    else {
      arrayVar = redefine_var(initName);
      Expr var_name = v.get_var_expr(dp);
      Expr schedule_id = sharedTerms[vname]->back();
      Expr previous_id = dp->mk_sub2(schedule_id, dp->mk_int_num(1), true);
      old_arrayVar = dp->mk_unary_app(var_name, previous_id);
    }

    // cout << "redirecting cout..." << endl;
    ofstream fileOutputStream;
    fileOutputStream.open ("trans.txt", ios_base::out | ios_base::binary);
    streambuf* sbuf = cout.rdbuf();
    cout.rdbuf(fileOutputStream.rdbuf());

    // (define b::bool (= f` (update f (pos_1 ... pos_n) new-val)))

    stringstream buf;
    
    // the variables in each MPI rank are distributed and are shared by threads during omp statement, edited by Hongyi. so this part need to built in block for MPI ranks.

    if (v.is_shared)
      buf <<  "#" << IN_PROC (vname, sharedVars[vname], thread_id);
    else
      {
       
       buf << "#" << IN_PROC (vname, localVars[vname], thread_id);
      }
     string name = buf.str();

    cout << "(define " << name << "::bool (= ";
    dp->pp_expr(arrayVar);
    cout << " (update ";
    dp->pp_expr(old_arrayVar);
    cout << " (";
    dp->pp_expr(args_stack.top());
    args_stack.pop();
    while (!args_stack.empty()) {
      Expr e = args_stack.top();
      args_stack.pop();
      cout << " ";
      dp->pp_expr(e);
    }
    cout << ") ";
    dp->pp_expr(value);
    cout << ")))" << endl;

    cout.rdbuf(sbuf);

    string line;
    ifstream infile ("trans.txt");
    while(!infile.eof()) {
      string x;
      getline (infile,x);
      line = line + x;
    }
    infile.close(); 
    cout << line << endl;
    dp->read_cmd(line);
    // debug information for current encoding   
    cout << "has error? " << yicesl_get_last_error_message() << endl;
    // cout << "name: " << name << endl;
    // ROSE_ASSERT(yicesl_get_last_error_message() == NULL);  // make sure the definition is in the context
      // cout << "error " << yicesl_get_last_error_message();

    Expr b = dp->get_var_from_name(name);
    return b;
  }

  else 
    return default_trans;
}


Expr ModelPF::buildTrans(SgNode* sgNode, SgNode* refExp, Expr& lhs) {

  Expr trans = default_trans;
  // SgVarRefExp* varRefExp = getRefVar(refExp);
  SgInitializedName* initName = getRefVar(refExp);

  if (isSgUnaryOp(sgNode)) {
    SgUnaryOp* unary = isSgUnaryOp(sgNode);
    switch (unary->variantT()) {
    case V_SgPlusPlusOp:
    case V_SgMinusMinusOp: {
      Expr e = convertExp(unary->get_operand());
      Expr var = redefine_var(initName);

      /*
      string v = initName->get_name().str();
      if (isSharedVar(v))
	lhs = makeGlobalVarIntactTrans(v);
      */

      switch (unary->variantT()) {
      case V_SgPlusPlusOp:
	trans = dp->mk_eq(var, dp->mk_sum2(e, dp->mk_num(1)));
	break;
      case V_SgMinusMinusOp: 
	trans = dp->mk_eq(var, dp->mk_sub2(e, dp->mk_num(1)));
	break;
      default:
	;
      }
    }
    default:
      ;
    }
  }
  else if (isSgBinaryOp(sgNode)) {
    SgBinaryOp* binary = isSgBinaryOp(sgNode);
    bool need_redefined = true;

    switch(binary->variantT()) {
    case V_SgAssignOp:
      need_redefined = false;
    case V_SgMinusAssignOp:
    case V_SgPlusAssignOp: 
    case V_SgMultAssignOp:
    case V_SgLshiftAssignOp:
    case V_SgRshiftAssignOp:
    case V_SgAndAssignOp: {

      Expr e = convertExp(binary->get_rhs_operand());

      string v = initName->get_name().str();

      Expr old_var;
      if (need_redefined) {
	cout << "Getting the old variable" << endl;
	old_var = convertExp(binary->get_lhs_operand());  // an operatorAssign operator
      }

      switch(binary->variantT()) {
      case V_SgAssignOp:
	trans = make_update_array_trans(refExp, e);
	break;
      case V_SgMinusAssignOp:
	trans = make_update_array_trans(refExp, dp->mk_sub2(old_var, e));
	break;
      case V_SgPlusAssignOp: 
	trans = make_update_array_trans(refExp, dp->mk_sum2(old_var, e));
	break;
      case V_SgMultAssignOp:
	trans = make_update_array_trans(refExp, dp->mk_mul2(old_var, e));
	break;
      case V_SgLshiftAssignOp:
	trans = make_update_array_trans(refExp, dp->mk_bv_shift_left(old_var, get_num_value(binary->get_rhs_operand())));
	break;
      case V_SgRshiftAssignOp:
	trans = make_update_array_trans(refExp, dp->mk_bv_shift_right(old_var, get_num_value(binary->get_rhs_operand())));
      case V_SgAndAssignOp:
	trans = make_update_array_trans(refExp, dp->mk_and2(old_var, e));
	break;	
	
      default:
	;
      }
      
      if (isSharedVar (v)) {
	lhs = makeGlobalVarIntactTrans(v);     // compliment the other branch
      }
    }
    default:
      ;
    }
  }

  else if (isSgAssignInitializer(sgNode)) {
    SgAssignInitializer* initializer = isSgAssignInitializer(sgNode);
    Expr e = convertExp(initializer->get_operand());
    string v = initName->get_name().str();

    trans = make_update_array_trans(refExp, e);
    cout << "trans = ";
    dp->pp_expr(trans);
    cout << endl;

    if (isSharedVar (v)) 
      lhs = makeGlobalVarIntactTrans(v);
  }

  return trans;
}

