#include "modelPF.h"

#define NOP "NOP";
#define NULL_VAR "??";

using namespace std;

/********************************************************************/

ostream &operator<<(ostream &stream, Variable var)
{
  stream << var.name;
  for (int i = 0; i < var.dim; i++) 
    stream << '[' << ']';
  return stream;
}

/********************************************************************/
//! this part shoud call from pragma omp block definitons.
// edited by Hongyi on 11/13/2014 in ompcheckerinterfac.h
  Expr ModelPF::mk_local_var(string name, VarNumMap& varMap) {
    stringstream out;
    VarNumMap::iterator ii = varMap.find(name);
    if (ii != varMap.end())
      {
      std::cout<<"testing name varMap[name] and thread id"<<name<<", "<<varMap[name]<<" , "<<thread_id<<std::endl;
      // ssa subsets
      out << IN_PROC(name, varMap[name], thread_id);
      }
     else
      out << IN_PROC(name, 0, thread_id);
    Variable v = lookupVars[name];
    return dp->mk_var(out.str(), v.get_var_type(dp));
  }

  // build v[i] = v[i-1 ]
  Expr ModelPF::mk_unchanged_trans(Expr var, Expr schedule_id) {
    // schedule ID - 1 
    
    Expr previous_id = dp->mk_sub2(schedule_id, dp->mk_int_num(1), true);
    Expr new_var = dp->mk_unary_app(var, schedule_id);
    Expr old_var = dp->mk_unary_app(var, previous_id);
    return dp->mk_eq(new_var, old_var);
  }


/********************************************************************/

void ModelPF::add_constraint(Expr e) {
  if (e != default_trans)
    constraints.push_back(e);
}

Expr ModelPF::sum_constraints(ExprList& constraints)  // the conjunction of all constraints in the list
{
  int size = constraints.size();
  Expr args[size];
 
  if (size == 0) 
    return dp->mk_true();
  else if (size == 1)
    return *(constraints.begin());
  else {
    int index = 0;
    for (list<Expr>::iterator i = constraints.begin(); i != constraints.end(); i++) 
      {
	args[index++] = (*i);
      }
    return dp->mk_and(args, size);
  }
}

Expr ModelPF::sum_constraints(stack<Expr> constraints)  // the conjunction of all constraints in the list
{
  int size = constraints.size();
  Expr args[size];
 
  if (size == 0) 
    return dp->mk_true();
  else if (size == 1)
    return constraints.top();

  int index = 0;
  while (!constraints.empty()) {
    args[index++] = constraints.top();
    constraints.pop();
  }
  return dp->mk_and(args, size);
}

void ModelPF::print_constraints(ExprList& constraints) {
  list<Expr>::iterator i;
  
  for (i = constraints.begin(); i != constraints.end(); i++) {
    dp->pp_expr(*i);
    cout << endl;
  }
}


/*****************************************************
 *****************************************************/

string formatOP (SgNode* sgNode)
{
  if (isSgUnaryOp(sgNode)) {
    SgUnaryOp* unary = isSgUnaryOp(sgNode);
    switch (unary->variantT()) {
    case V_SgPlusPlusOp:
      return "++";
    case V_SgMinusMinusOp: 
      return "--";
    default:
      return NOP;
    }
  }

  else if (isSgBinaryOp(sgNode)) {
    SgBinaryOp* binary = isSgBinaryOp(sgNode);
    switch(binary->variantT()) {
    case V_SgArrowExp:
      return "->";
    case V_SgDotExp:
      return ".";
    case V_SgDotStarOp:
      return ".*";
    case V_SgArrowStarOp:
      return "->*";
    case V_SgEqualityOp:
      return "==";
    case V_SgLessThanOp:
      return "<";
    case V_SgGreaterThanOp:
      return ">";
    case V_SgNotEqualOp:
      return "<>";
    case V_SgLessOrEqualOp:
      return "<=";
    case V_SgGreaterOrEqualOp:
      return ">=";
    case V_SgAddOp:
      return "+";
    case V_SgSubtractOp:
      return "-";
    case V_SgMultiplyOp:
	return "*";
    case V_SgDivideOp:
    case V_SgIntegerDivideOp:
      return "/";
    case V_SgModOp:
      return "%";
    case V_SgAndOp:
      return "&&";
    case V_SgOrOp:
      return "||";
    case V_SgBitXorOp:
      return "^";
    case V_SgBitAndOp:
      return "&";
    case V_SgBitOrOp:
      return "|";
	// SgCommaOpExp, 
    case V_SgLshiftOp:
      return "<<";
    case V_SgRshiftOp:
      return ">>";
    case V_SgPntrArrRefExp:
      return "[]";
      // SgScopeOp,
    case V_SgAssignOp:
      return ":=";
    case V_SgModAssignOp:
      return "%=";
    case V_SgDivAssignOp:
      return "/=";
    case V_SgMultAssignOp:
	return "*=";
    case V_SgLshiftAssignOp:
      return "<<=";
    case V_SgRshiftAssignOp:
	return ">>=";
    case V_SgXorAssignOp:
      return "|=";
    case V_SgAndAssignOp:
      return "&=";
    case V_SgMinusAssignOp:
      return "-=";
    case V_SgPlusAssignOp:
      return "+=";
    default: 
      return NOP;
    }
  }

  else 
    return NOP;
}


// note: use sgNode->variantT() rather than isSgVarRefExp(sgNode) won't work 
// since ROSE seems not to put the right type on the SgNode.
string formatExp(SgNode* sgNode)
{
  if (isSgVarRefExp(sgNode)) {   // variables
    SgVarRefExp* varRefExp = isSgVarRefExp(sgNode);
    SgInitializedName* initName = varRefExp->get_symbol()->get_declaration();
    return initName->get_name().str();
  }
  
  else if (isSgValueExp(sgNode)) { // constants
    SgValueExp* varRef = isSgValueExp(sgNode);
    // SgName name = varRef->get_symbol()->get_name();
    // return varRef->class_name();
    stringstream out;
    
    switch(varRef->variantT()) {
    case V_SgIntVal:
      out << isSgIntVal(varRef)->get_value();
      return out.str();
    case V_SgFloatVal:
      out << isSgFloatVal(varRef)->get_value();
      return out.str();
    case V_SgDoubleVal:
      out << isSgDoubleVal(varRef)->get_value();
      return out.str();
    default:
      return "0";
    }
  }

  else if (isSgUnaryOp(sgNode)) {
    SgUnaryOp* unary = isSgUnaryOp(sgNode);
    SgExpression* l_expr = unary->get_operand();
    return formatExp(l_expr) + formatOP(sgNode);
  }

  else if (isSgBinaryOp(sgNode)) {
    SgBinaryOp* binary = isSgBinaryOp(sgNode);
    SgExpression* l_expr = binary->get_lhs_operand();
    SgExpression* r_expr = binary->get_rhs_operand();
    return formatExp(l_expr) + formatOP(sgNode) + formatExp(r_expr);
  }

  else if (isSgAssignInitializer(sgNode)) {
    SgAssignInitializer* initializer = isSgAssignInitializer(sgNode);
    SgNode* name_expr = initializer->get_parent();

    if (isSgInitializedName(name_expr)) {
      SgInitializedName* initName =isSgInitializedName(name_expr);
      return initName->get_name().str() + string(":=") + 
	formatExp(initializer->get_operand()); 

    } else if (isSgExprListExp(name_expr)) {
    }

    /*
      SgExpression* l_expr = binary->get_lhs_operand();
      SgExpression* r_expr = binary->get_rhs_operand();
      return formatExp(l_expr) + formatOP(sgNode) + formatExp(r_expr);
    */
    return NOP;
    }

  /*
  case V_SgFunctionCallExp:
  {
    SgFunctionCallExp* fcallExp = isSgFunctionCallExp(sgNode);
    SgExprListExp* exprList = fcallExp->get_args();
    SgExpressionPtrList& list = exprList->get_expressions();

    SgExpressionPtrList::iterator i;
    string s = "";
    for (i = list.begin(); i != list.end(); ++i) {
      SgExpression* expr = *i;
      s = s + formatExp(expr);
    }
    return fcallExp + "(" + s + ")";
  }
  */
  else 
    return NULL_VAR;
}


/*
void visit(SgNode* sgNode)
{
  if (isSgVarRefExp(sgNode)) {   // variables
    SgVarRefExp* varRefExp = isSgVarRefExp(sgNode);
    SgInitializedName* initName = varRefExp->get_symbol()->get_declaration();
    return initName->get_qualified_name().str();
  }
  
  else if (isSgValueExp(sgNode)) { // constants
    SgValueExp* varRef = isSgValueExp(sgNode);
    // SgName name = varRef->get_symbol()->get_name();
    // return varRef->class_name();
    stringstream out;
    
    switch(varRef->variantT()) {
    case V_SgIntVal:
      out << isSgIntVal(varRef)->get_value();
      return out.str();
    case V_SgFloatVal:
      out << isSgFloatVal(varRef)->get_value();
      return out.str();
    case V_SgDoubleVal:
      out << isSgDoubleVal(varRef)->get_value();
      return out.str();
    default:
      return "0";
    }
  }

  else if (isSgUnaryOp(sgNode)) {
    SgUnaryOp* unary = isSgUnaryOp(sgNode);
    SgExpression* l_expr = unary->get_operand();
    return formatExp(l_expr) + formatOP(sgNode);
  }

  else if (isSgBinaryOp(sgNode)) {
    SgBinaryOp* binary = isSgBinaryOp(sgNode);
    SgExpression* l_expr = binary->get_lhs_operand();
    SgExpression* r_expr = binary->get_rhs_operand();
    return formatExp(l_expr) + formatOP(sgNode) + formatExp(r_expr);
  }

  else if (isSgAssignInitializer(sgNode)) {
    SgAssignInitializer* initializer = isSgAssignInitializer(sgNode);
    SgNode* name_expr = initializer->get_parent();

    if (isSgInitializedName(name_expr)) {
      SgInitializedName* initName =isSgInitializedName(name_expr);
      return initName->get_qualified_name().str() + string(":=") + 
	formatExp(initializer->get_operand()); 

    } else if (isSgExprListExp(name_expr)) {
    }

    return NOP;
    }

  case V_SgFunctionCallExp:
  {
    SgFunctionCallExp* fcallExp = isSgFunctionCallExp(sgNode);
    SgExprListExp* exprList = fcallExp->get_args();
    SgExpressionPtrList& list = exprList->get_expressions();

    SgExpressionPtrList::iterator i;
    string s = "";
    for (i = list.begin(); i != list.end(); ++i) {
      SgExpression* expr = *i;
      s = s + formatExp(expr);
    }
    return fcallExp + "(" + s + ")";
  }

  else 
    return NULL_VAR;
}
*/
