#include "ompchecker.h"
#include <time.h>
//! last edited by Hongyi on September 16th, 2014
// for loop unrolling
#include "rose.h"
#include <AstInterface_ROSE.h>
#include "LoopTransformInterface.h"
#include "CommandOptions.h"
#include "LoopUnroll.h"
#include "util.h"

//mpi suppport is implemented in model.h
#include <iostream>
#include <algorithm>
#include <numeric>
using namespace std;

// no needed only interested in statements inside
// preprocessing for loop unrolling bound

/*************************************/


/**************************************************************************************
 *************************************************************************************/




/*************************************/
using std::max;

std::vector<SgFunctionCallExp*> funcList;


//mpi fucntioncalls support 

static struct MpiGatherFunction
{
 const char* name;
 int root;

}mpiGatherFunction = {"MPI_Gather", 6};

static struct MpiReduceFunction
{
 
 const char* name;
 int root;

}mpiReduceFunction = {"MPI_Reduce", 5};




static struct MpiSendFunction
{
 const char* name;
 int sourcePos;
 int tagPos;
}mpiSendFunctions[] = {
 {"MPI_Send", 3,4},
 {"MPI_Isend", 3,4},
 {"MPI_Send_init",3,4},
 {"MPI_Sendrecv", 8, 9},
 {"MPI_Sendrecv_replace", 5, 6}
};

static struct MpiRecvFunction {
  const char *name;
  // The position of the source and tag arguments;
  int sourcePos;
  int tagPos;
} mpiRecvFunctions[] = {
      {"MPI_Recv", 3, 4}
     ,{"MPI_Irecv", 3, 4}
     ,{"MPI_Recv_init", 3, 4}
     ,{"MPI_Sendrecv", 8, 9}
     ,{"MPI_Sendrecv_replace", 5, 6}
};

MpiDeterminism combine(MpiDeterminism d1, MpiDeterminism d2)
{
  MpiDeterminism d;
  d.source = max(d1.source, d2.source);
  d.tag = max(d1.tag, d2.tag);
  d.functions = max(d1.functions, d2.functions);
  return d;
}

MpiDeterminism MpiDeterminismAnalysis::evaluateSynthesizedAttribute(SgNode *n, SynthesizedAttributesList childAttribs)
{
  MpiDeterminism children = std::accumulate(childAttribs.begin(), childAttribs.end(), MpiDeterminism(), combine);

  SgFunctionCallExp *fncall = dynamic_cast<SgFunctionCallExp *>(n);

  return fncall ? combine(checkCall(fncall), children) : children;
}

void MpiDeterminismAnalysis::bump(DeterminismState &target, DeterminismState value)
{
  target = max(target, value);
}

MpiDeterminism MpiDeterminismAnalysis::checkCall(SgFunctionCallExp *fncall)
{
  MpiDeterminism d;
  SgFunctionDeclaration *fn = fncall->getAssociatedFunctionDeclaration();
  SgName fnname = fn->get_name();

  for (int i = 0; i < sizeof(mpiRecvFunctions)/sizeof(mpiRecvFunctions[0]); ++i) {
    if (fnname == mpiRecvFunctions[i].name) {
      //! build a vector to store all functionCall
     // vector<SgFunctionCallExp*> funcList;
      funcList.push_back(fncall);
      std::cout<<"the current size of functioncall is "<<funcList.size()<<std::endl;    
      SgExpressionPtrList args = fncall->get_args()->get_expressions();
      
      //! edited by Hongyi to fetch all variable names
      
      SgExpressionPtrList::iterator arg = args.begin();
      for(; arg!=args.end(); arg++)
      {
       std::cout<<"arg =  "<<(*arg)->unparseToString()<<std::endl;
        
      }
      //end     
      std::cout<<"we have found argument list "<<std::endl;
      // Source argument testing
      SgExpression *sourceArg = args[mpiRecvFunctions[i].sourcePos];
      SgVarRefExp *sourceVar = dynamic_cast<SgVarRefExp *>(sourceArg);
      SgIntVal *sourceInt = dynamic_cast<SgIntVal *>(sourceArg);
      if (sourceVar) {
	SgName name = sourceVar->get_symbol()->get_name();
	bump(d.source, name == "MPI_ANY_SOURCE" ? NONDETERMINISTIC : QUESTIONABLE);
      } else if (sourceInt) {
	if (sourceInt->get_value() < 0) {
	  std::cerr << "Negative literal source rank seen - not using the hacked mpi.h?" << std::endl;
	  bump(d.source, QUESTIONABLE);
	}
      } else {
	bump(d.source, QUESTIONABLE);
      }

      // Tag argument testing
      SgExpression *tagArg = args[mpiRecvFunctions[i].tagPos];
      SgVarRefExp *tagVar = dynamic_cast<SgVarRefExp *>(tagArg);
      SgIntVal *tagInt = dynamic_cast<SgIntVal *>(tagArg);
      if (tagVar) {
	SgName name = tagVar->get_symbol()->get_name();
	bump(d.tag, name == "MPI_ANY_TAG" ? NONDETERMINISTIC : QUESTIONABLE);
      } else if (tagInt) {
	if (tagInt->get_value() < 0) {
	  std::cerr << "Negative literal tag seen - not using the hacked mpi.h?" << std::endl;
	  bump(d.tag, QUESTIONABLE);
	}
      } else {
	bump(d.tag, QUESTIONABLE);
      }
      
    }
  }

  if (fnname == "MPI_Waitany" || fnname == "MPI_Testany" || fnname == "MPI_Waitsome" || fnname == "MPI_Testsome")
    bump(d.functions, NONDETERMINISTIC);

  return d;
}

/**************************************/

//edited by Hongyi for testing the fucntionality of MPI arguments 

using std::cerr;
using std::cout;
using std::endl;

DeterminismState getExpectation(SgNode *ast, const char *varName)
{
  SgName name(varName);

  Rose_STL_Container<SgNode*> sdNodes = NodeQuery::querySubTree(ast, &name, NodeQuery::VariableDeclarationFromName);
  if (sdNodes.size() != 1) {
    cerr << "Didn't find target variable " << varName << " in list of size " << sdNodes.size() << endl;

    for (Rose_STL_Container<SgNode*>::iterator i = sdNodes.begin(); i != sdNodes.end(); ++i)
      cerr << "\t" << (*(isSgVariableDeclaration(*i)->get_variables().begin()))->get_name().str() << endl;

    return QUESTIONABLE;
  }

  SgNode *nSd = *(sdNodes.begin());
  SgVariableDeclaration *vdSd = dynamic_cast<SgVariableDeclaration *>(nSd);
  if (!vdSd) {
    cerr << "Node wasn't a variable declaration" << endl;
    return QUESTIONABLE;
  }
 
   SgInitializedName *inSd = vdSd->get_decl_item(name);
  SgAssignInitializer *aiSd = dynamic_cast<SgAssignInitializer*>(inSd->get_initializer());
  if (!aiSd) {
    cerr << "Couldn't pull an assignment initializer out" << endl;
    return QUESTIONABLE;
  }

  SgIntVal *ivSd = dynamic_cast<SgIntVal*>(aiSd->get_operand());
  if (!ivSd) {
    cerr << "Assignment wasn't an intval" << endl;
    return QUESTIONABLE;
  }

  int value = ivSd->get_value();
  return value ? DETERMINISTIC : NONDETERMINISTIC;
}

const char* strFromDet(DeterminismState d)
{
  switch (d) {
  case DETERMINISTIC: return "deterministic";
  case QUESTIONABLE: return "unclear";
  case NONDETERMINISTIC: return "NONdeterminisitic";
  }
}

void report(DeterminismState actual, DeterminismState expected, const char *kind, int &incorrect, int &imprecise)
{
  cout << "\t" << strFromDet(actual) << " (expected " << strFromDet(expected) << ") with respect to receive " << kind << endl;

  if (QUESTIONABLE == actual && QUESTIONABLE != expected)
    imprecise++;
  else if (actual != expected)
    incorrect++;
}











/******************************************




 ******************************************/

void Model::run() {
  
  if (DEBUG_MODE) 
    std::cout << "START: Analysis " <<  (DEBUG_MODE ? "True" : "False") << endl;
   
  nrOfNodesVisited = 0;

  // Traverse through each FunctionDefinition
  Rose_STL_Container<SgNode*> functions = NodeQuery::querySubTree(project, V_SgFunctionDefinition);   
  for (Rose_STL_Container<SgNode*>::const_iterator i = functions.begin(); i != functions.end(); ++i) {
    SgFunctionDefinition* proc = isSgFunctionDefinition(*i);
    SgFunctionDeclaration* funcDeclaration = proc->get_declaration();
    string funcName = funcDeclaration->get_name().str();
    // test by Hongyi for several functions inline extern or global
 
    //previous we consider global or local to distingush them
    // but currently, we consider general function instead
    //these functions also being considered as shared or global. 
    if (check_func == NULL && (is_check_func(funcDeclaration) || is_global_func(funcDeclaration)))
      check_func = proc;
    else
     { 
      general_func = proc;
     }
 // } // functioncall

  for (int i = 0; i < num_of_threads; i++) {
    AnalysisPF* analysis = new AnalysisPF(&dp, DEBUG_MODE);
    funcs.push_back(analysis);

    analysis->module_SSA->num_of_threads = num_of_threads;
    analysis->module_SSA->thread_id = i;
    analysis->project = project;
    //edited by Hongyi change the check_func to general_func;
    Expr e;
    if(check_func==NULL)
     {
       e = analysis->start_traversal_of_functions(general_func);
     }
     else
     {
     //edited by Hongyi 
     // we mostly care about function call in parallel region. 

     e = analysis->start_traversal_of_functions(check_func);
     }
    transitions.push_back(e);     // transitions
    genConstraints.push_back(&(funcs[i]->module_SSA->constraints)); // non path sensitive constraitns 
    genSharedTerms.push_back(&(funcs[i]->module_SSA->sharedTerms)); // schedule ids of shared variables
    genProperties.push_back(&(funcs[i]->module_SSA->properties)); 
    // For data race checking, properties have race constraints.

    if (barriers == NULL)
      barriers = &(funcs[i]->module_SSA->barriers);
  }

 }
  // compose_transistions();
  Expr e = compose_constraints();
  dp.pp_expr(e);
  cout << endl;
  cout << "check the expression:" << endl;
  check(e);

  // release memory
  for (int i = 0; i < num_of_threads; i++)
    delete funcs[i];
  cout << "Done !!!!!!!!!!!!!!!" << endl;
}

/********************************************************
 Check the entire transition along with the properties.
********************************************************/

void Model::check(Expr e) {

  // execution time
  clock_t start, end;
  double dif;

  start = clock();

  dp.DP_assert(e);
 // dp.dump_context();

  switch(dp.check()) {
  case l_true: {
    printf("satisfiable\n");
    DP_model m = dp.get_model();
    dp.display_model(m);
    //edited by Hongyi for cost
    long long cost  = dp.get_cost(m);
    printf("current cost is %ld \n", cost);  
   
    break;  
  }
  case l_false:
    printf("unsatisfiable\n");
    break;
  case l_undef:
    printf("unknown\n");
    break;
  }

  end = clock();

  dif = ((double) (end - start)) / CLOCKS_PER_SEC;
  cout << "The checking time is: " << dif << endl;
}

/******************************************
 Build the entire transition.
 ******************************************/

Expr Model::mk_constraints()  // the conjunction of all constraints in the list
{
  int size = num_of_threads * (genConstraints[0]->size() + 1);
  // cout << "size = " << size << endl; 
  Expr args[size];
  
  int index = 0;
  int i = 0;
  for (vector<list<Expr>*>::iterator ii = genConstraints.begin(); ii != genConstraints.end(); ii++) { // threads
    args[index++] = transitions[i++];
    for(list<Expr>::iterator jj = (*ii)->begin(); jj != (*ii)->end(); jj++) {
      args[index++] = *jj;
    }
  }
  // put constraints together with and symbol
  return dp.mk_and(args, index);
}

Expr Model::mk_properties()  // the conjunction of all constraints in the list
{
  int size = num_of_threads * (*genProperties.begin())->size();
  // cout << "size = " << size << endl; 
  Expr args[size];
  
  int index = 0;
  vector<list<Expr>*>::iterator ii;
  for(ii = genProperties.begin(); ii != genProperties.end(); ii++) 
    for(list<Expr>::iterator jj = (*ii)->begin(); jj != (*ii)->end(); jj++) {
      args[index++] = *jj;
    }
  return dp.mk_and(args, index);
}

// make disequalities through function facility
Expr Model::mk_global_order_constraints() {
  vector<Expr> args;

  Type ty = dp.mk_nat_type();
  Type fty = dp.mk_unary_function_type(ty, ty);

  ModelPF::VarListMap *vars = genSharedTerms[0];
  Expr results[vars->size()];
  int v_count = 0;

  for (ModelPF::VarListMap::iterator ii = vars->begin(); ii != vars->end(); ii++) { // for all variables
    string v = ii->first;
    cout << "v = " << v << endl;
    args.clear();
    for (vector<ModelPF::VarListMap*>::iterator jj = genSharedTerms.begin(); 
	 jj != genSharedTerms.end(); jj++) { // for all threads
      list<Expr>* ids = (*(*jj))[v];
      for (list<Expr>::iterator kk = ids->begin(); kk != ids->end(); kk++) {
	args.push_back(*kk);  // rank[sn] = index
      }
    }
    list<Expr>& bars = (*barriers)[v];  
    for(list<Expr>::iterator ll = bars.begin(); ll != bars.end(); ll++)  // for barriers
      args.push_back(*ll);
    
    int size = args.size();
    varScheduleLens[v] = size;    // for post condition checking

    Expr size_term = dp.mk_int_num(size);
    Expr args1[size*2];
    int index = 0;
    int id_count = 0;
    Expr rank_term = dp.mk_var("rank_"+v, fty);
 
    for (vector<Expr>::iterator kk = args.begin(); kk != args.end(); kk++) {
      args1[index++] = dp.mk_eq(dp.mk_unary_app(rank_term, *kk), dp.mk_int_num(id_count++));
      args1[index++] = dp.mk_lt(*kk, size_term, true);
    }
    results[v_count++] = dp.mk_and(args1, index);

    // dp.pp_expr(results[v_count-1]);
    // cout << endl;
  }
 
  return dp.mk_and(results, v_count);
}


Expr Model::mk_condition(bool isPost) {               // check the post condition

  if (isPost && guarantee_func == NULL) 
    return dp.t_expr;
  else if (!isPost && assume_func == NULL)
    return dp.t_expr;

  map<string,Expr> vmap;
  SgFunctionDeclaration* funcDeclaration;
  SgBasicBlock* body;

  if (isPost) {
    funcDeclaration = guarantee_func->get_declaration();
    body = guarantee_func->get_body();

    for(map<string, int>::iterator ii = varScheduleLens.begin(); ii != varScheduleLens.end(); ii++) {
      vmap[ii->first] = dp.mk_int_num(ii->second - 1);
    }
  }
  else {
    funcDeclaration = assume_func->get_declaration();
     body = assume_func->get_body();
    for(map<string, int>::iterator ii = varScheduleLens.begin(); ii != varScheduleLens.end(); ii++) {
      vmap[ii->first] = dp.mk_int_num(-1);
    }
  }

  Expr args[50];
  int index = 0;

  Rose_STL_Container<SgNode*> functions = NodeQuery::querySubTree(body, V_SgFunctionCallExp);   
  for (Rose_STL_Container<SgNode*>::const_iterator i = functions.begin(); i != functions.end(); i++) {
    SgFunctionCallExp *fcallExp = isSgFunctionCallExp(*i);
    SgFunctionSymbol *symbol = fcallExp->getAssociatedFunctionSymbol();
    if (strcmp(symbol->get_name().str(),"assert") == 0) {
      SgExprListExp* exprList = fcallExp->get_args();
      SgExpressionPtrList& list = exprList->get_expressions();
      for (SgExpressionPtrList::iterator j = list.begin(); j != list.end(); j++) {
	SgExpression* expr = *j;
	args[index++] = funcs[0]->module_SSA->convertExp(expr, &vmap);
      }
    }
  }
  Expr e = dp.mk_and(args, index);
  return e;
}

Expr Model::compose_constraints() {

  cout << "Constraints: " << endl;
  Expr cons = mk_constraints();
  dp.pp_expr(cons);
  cout << endl;

  cout << "Global Order: " << endl;
  Expr global_order = Config::use_POR ? mk_POR() : mk_global_order_constraints();
  dp.pp_expr(global_order);
  cout << endl;

  cout << "Checked Properties: " << endl;
  Expr props = mk_properties();
  dp.pp_expr(props);
  cout << endl;

  //commented by Hongyi, always consider post conditions is enough, so set true for it

  cout << "Pre and Post Condition: " << endl;
  Expr pre = mk_condition(false);
  Expr post = mk_condition(true);

  // check all unsure functions assume \land not(guarantee without data races) 
  Expr cond = dp.mk_true();
  if (post != cond) {
    cond = pre == dp.t_expr ? dp.mk_not(post) : dp.mk_and2(pre, dp.mk_not(post));     // pre /\ not post
  }
  dp.pp_expr(cond);
  cout << endl;

   Expr args[] = {cons, global_order, props, cond};
   return dp.mk_and(args, 4);
}

string get_vname(DP& dp, Expr& e) {
  ofstream fileOutputStream;
  fileOutputStream.open ("v.txt");
  streambuf* sbuf = cout.rdbuf();
  cout.rdbuf(fileOutputStream.rdbuf());
  dp.pp_expr(e);
  cout.rdbuf(sbuf);
  
  string line;
  ifstream infile ("tmp.txt");
  getline (infile,line);
  infile.close();
  
  // cout << "name = " << line << endl;
  return line;
}

 // build Partial Order Reduction constraints list
Expr Model::mk_POR() {
  vector<Expr> args;

  ModelPF::VarListMap *vars = genSharedTerms[0];
  Expr results[vars->size()];
  int v_count = 0;

  for (ModelPF::VarListMap::iterator ii = vars->begin(); ii != vars->end(); ii++) { // for all variables

    string v = ii->first;
    cout << "variable: " << v << endl;
    args.clear();

    vector<list<Expr>::iterator> BI_iters;
    vector<list<Expr>*> SIds;
    int index = 0;

    for (vector<ModelPF::VarListMap*>::iterator jj = genSharedTerms.begin(); 
	 jj != genSharedTerms.end(); jj++) { // for all threads
      list<Expr>* ids = (*(*jj))[v];
      SIds.push_back(ids);
      BI_iters.push_back(ids->begin());
    }

    do {
      for (int i = 0; i < num_of_threads; i++) {

        // bar means barrier in Openmp
	while (BI_iters[i] != SIds[i]->end() && 
	       get_vname(dp, *(BI_iters[i])).find("bar") == string::npos) {
	  args.push_back(dp.mk_eq(*(BI_iters[i]), dp.mk_int_num(index++)));  // SId = index
	  BI_iters[i]++;
	}
      }

      if (BI_iters[0] != SIds[0]->end()) {
	args.push_back(dp.mk_eq(*(BI_iters[0]), dp.mk_int_num(index++)));
	for (int i = 0; i < num_of_threads; i++) {
	  BI_iters[i]++;  // skip the barrier
	}
      }
    }
    while (BI_iters[0] != SIds[0]->end());

    Expr args1[index];
    index = 0;
    for (vector<Expr>::iterator kk = args.begin(); kk != args.end(); kk++) {
      args1[index++] = *kk;
    }
    results[v_count++] = dp.mk_and(args1, index);

    // dp.pp_expr(results[v_count-1]);
    // cout << endl;
  }
 
  return dp.mk_and(results, v_count);
}


/*********************************************************************
 ********************************************************************/

int main( int argc, char * argv[]) {

  //defaut two threads
  int N = 2;
  int argc1 = argc;
  char** argv1 = argv;

  //! parse the command line list
  while(argc1--) {
    string arg = *argv1++;
    if (arg == "-ssa" || arg == "-SSA") {
      cout << "Use ssa" << endl;
      Config::use_POR = true; //parital order reduction
    }
    else if (arg[0] == '-' && (arg[1] == 't' || arg[1] == 'T')) {
      N = atoi(arg.substr(2, arg.length()).c_str());
      cout << "#threads = " << N << endl;
    }
  }

  // Config::use_ssa = true;

  SgProject* project = frontend(argc, argv);
  std::cout << ">>>> start analysis with OpenMP and MPI... " << endl;

  SgName hacked("MPI_HACKED_HEADER_INCLUDED");
  // check the mpi.h included in the soruce code analysis 

  Rose_STL_Container<SgNode*> vars = NodeQuery::querySubTree(project,&hacked, NodeQuery::VariableDeclarationFromName );
 
 if(vars.size() != 1)
  {
     cerr << "You must be using the hacked mpi.h that defines things nicely for this to work!" << endl;
    //  return 10;
     
  }


  //! check the expected result 
  MpiDeterminismAnalysis anaMPI;
  MpiDeterminism d = anaMPI.traverse(project);
  DeterminismState sourceExpection = getExpectation(project, "SOURCE_DETERMINISM");
  DeterminismState tagExpection = getExpectation(project, "TAG_DETERMINISM");
  DeterminismState functionExpection = getExpectation(project, "FUNCTION_DETERMINISM");
  
 //*****************************************************************//

  int incorrect = 0, imprecise = 0;

  cout << "Analysis finds that this program is" << endl;
  report(d.source, sourceExpection, "sources", incorrect, imprecise);
  report(d.tag, tagExpection, "tags", incorrect, imprecise);
  report(d.functions, functionExpection, "functions", incorrect, imprecise);

  cout << "Analysis was precise in " << 3 - incorrect - imprecise 
       << " cases, imprecise in " << imprecise 
       << " cases, and WRONG in " << incorrect << " cases." << endl;
   


  //*************************************************************//  
  
  Model* model = new Model(project, N);
  model->run();
  //finished encoding
  delete model;
  return 0;
}
