#ifndef __MY_MODEL__
#define __MY_MODEL__
#include "analysisPF.h"
#include <iostream>
#include <algorithm>
#include <numeric>
#include "rose.h"

//!edited by Hongyi, since openmp has similar local, global variable issues,
// so we match guarantte like atomic, single, master, omp section
// and kernel means variables defined in local
// assume is for global varibles 


//! fucntioncalls in parallel region


#define general_str ""
//! assume data race may happen inside
//! no data race happen in these functions


class Model  {
 private:
  Rose_STL_Container<SgNode*> functions;
  SgProject* project;

  bool mpiomp;
  bool DEBUG_MODE;
  static int sgNodeCounter;
  int nrOfNodesVisited;

  int num_of_threads;
  DP dp;

  vector < AnalysisPF* > funcs;
  
  SgFunctionDefinition* check_func;
  SgFunctionDefinition* assume_func;
  //check single thread region
  SgFunctionDefinition* guarantee_func;
  
 // edited by Hongyi for general function in c/c++ program
  SgFunctionDefinition* general_func;
 
  
  vector<Expr> transitions;                 // transitions
  vector<list<Expr>*> genConstraints;       // non path sensitive constraitns 
  vector< map<string,list<Expr>*>* > genSharedTerms;     // shared variables
  vector<list<Expr>*> genProperties;      // the constraitns for the properties to be checked 
  map<string, list<Expr> > *barriers;

  map<string, int> varScheduleLens;       // the schedule length of each shared variable 

  Expr mk_constraints();                  // transitions and non path sensitive constraitns 
  Expr mk_global_order_constraints();     // order the events across the threads
  Expr mk_POR();                          // apply the POR algorithm to generate orders
  Expr mk_properties();                   // check the properties
  Expr mk_condition(bool isPost = true);  // check the pre and post condition
  Expr compose_constraints();             // compose all constraints

 public:

  Model(SgProject* project_in, int numThreads = 2)
    {
      project = project_in;
      functions = NodeQuery::querySubTree(project, V_SgFunctionDefinition);  // the functions
      num_of_threads = numThreads;
      barriers = NULL;
      mpiomp = NULL;
      check_func = NULL;
     // assume_func = NULL;
     //  guarantee_func = NULL;
    };
  virtual ~Model() {
  }
  
  // AnalysisPF* start_traversal_of_functions(int thread_id);

  void run();
  void check(Expr e);

};



enum DeterminismState { DETERMINISTIC = 0, QUESTIONABLE = 1, NONDETERMINISTIC = 2 };

struct MpiDeterminism
{
  DeterminismState source, tag, functions;
  MpiDeterminism() : source(DETERMINISTIC), tag(DETERMINISTIC), functions(DETERMINISTIC) { }
};

class MpiDeterminismAnalysis : public AstBottomUpProcessing<MpiDeterminism>
{
public:
  MpiDeterminismAnalysis() { }
  MpiDeterminism evaluateSynthesizedAttribute(SgNode *n, SynthesizedAttributesList childAttribs);

private:
  void bump(DeterminismState &target, DeterminismState value);
  MpiDeterminism checkCall(SgFunctionCallExp *fncall);
};



#endif
